/*
 ArdBot full function demonstration
 Requires Arduino IDE version 0019 or later
 Requires PololuQTRSensors library
  (available from Pololu; see text)
 Requires MsTimer2 library (available
  from Arduino.cc libraries page)
*/

#include <Servo.h>
#include <PololuQTRSensors.h>   // Add to library
#include <MsTimer2.h>           // Add to library

Servo servoLeft;           // Left servo
Servo servoRight;          // Right servo
Servo servoTurret;         // Turret servo

// General member variables
#define LED 13             // LED pin
int mode = LOW;            // LOW = explore
int sonar = 0;
boolean turret_enable = false;
boolean mstimer = false;

// 2-sensor line following detector
#define lineLSense A3      // Analog pin A3
#define lineRSense A4      // Analog pin A4
int thresh = 400;          // Brightness threshold for line
int irReflectR = 0;        // Storage vars for line sensors
int irReflectL = 0;

// QTR-8 line following array setup
#define NUM_SENSORS   5     // Number of sensors
#define TIMEOUT       2500  // Maximum timeout per sensor

// QTR-8 connected to digital pins 4, 5, 6, 11, and 12
PololuQTRSensorsRC qtrrc((unsigned char[]) {4, 5, 6, 11, 12},
  NUM_SENSORS, TIMEOUT, QTR_NO_EMITTER_PIN);
unsigned int sensorValues[NUM_SENSORS];
#define VARIANCE 300       // Variance from center (2000)

// CdS cell eyes
#define BRIGHT 800           // Brightness threshold
#define cdsLeft A0           // Left photocell
#define cdsRight A1          // Right photocell
int eyeR = 0;                // Storage vars for CdS cells
int eyeL = 0;

// Sonar and IR
int duration;                 // Stores duration of pulse
int distance;                 // Stores distance
int averaging;                // Stores average of distance
#define srfPin 7              // SRF05 connected to pin D7
#define T_DELAY 1000          // Delay for turret resettle
#define NEAR 225              // IR proximity threshold

// Bumper switches
#define bumpLeft 2            // Switch to pin D2
#define bumpRight 3           // Switch to pin D3
int pbLeft = 0;               // Storage vars for bumpers
int pbRight = 0;

void setup() {
  servoLeft.attach(10);       // Left servo pin D10
  servoRight.attach(9);       // Right servo pin D9
  servoTurret.attach(8);      // Turret servo pin D8  
  
  if (digitalRead(A2) == HIGH) { // Read mode jumper
    mode = HIGH;                 // Line follow mode
    delay(500);
    pinMode(13, OUTPUT);
    digitalWrite(13, HIGH);     // Show calibrating
    for (int i=0; i<400; i++) {
      qtrrc.calibrate();        // Read sensors to calibrate
    }
    digitalWrite(13, LOW);      // Calibrate completed
    delay(1000);
  } else {                      // Explore mode
    // Set pin modes
    pinMode(bumpLeft, INPUT);  
    pinMode(bumpRight, INPUT);  
    pinMode(LED, OUTPUT);

    // Set up button interrupts
    attachInterrupt(0, hitLeft, RISING);
    attachInterrupt(1, hitRight, RISING); 

    //Set timer interrupt
    MsTimer2::set(10000, turret_scan);  // Every 10 seconds
    MsTimer2::start();
  }
}

void loop() {
  if (mode)
    mode_line_follow();
  else
    mode_explore();
}

void mode_explore() {
  if (mstimer) {               // If timer has triggered
    if (turret_enable) {       // If timer event enabled
      turret_enable = false;
      stopRobot();             // Momentarily stop robot
      turretR();               // Turn turret to right and
      if (irRead()) {          //  take a reading. Turn robot
        turnLeft();            //  to left if an object is detected
        delay(300);
      } else {
        turretL();             // Do same with turret to left
        if (irRead()) {
          turnRight();
          delay(300);
        }
      }
      turretC();              // Recentet turret
    }
  }
  forward();                  // Drive forward and reset timer
  mstimer = false;
  turret_enable = true;
  
  showLED();                  // Show LED indicator
  
  // Left bumper hit
  if (pbLeft == HIGH) {
    turret_enable = false;
    reverse();
    delay(500); 
    turnRight();
    delay(1500);
    pbLeft = LOW;
    turret_enable = true;    
  }
  
  // Right bumper hit
  if (pbRight == HIGH) {
    turret_enable = false;    
    reverse();
    delay(500); 
    turnLeft();
    delay(1500);
    pbRight = LOW;
    turret_enable = true;    
  }
  
  // Check for bright lights ahead
  bright_lights();

  // Read sonar to detect oncoming collision
  sonar = sonarRead();
  if(sonar < 4) {      // Object ~3" or closer
    forwardSlow();
  }
  if(sonar < 2) {      // Object ~1" or closer
    stopRobot();
    delay(250);
    reverse();
    delay(500);
    turnRight();
    delay(750);
    forward();
  }
}

// Shy away from bright lights if light on cells
//  exceeds BRIGHT value
void bright_lights() {
  eyeL = analogRead(cdsLeft);  
  eyeR = analogRead(cdsRight);
  if (eyeL > BRIGHT && eyeR > BRIGHT) {
    for (int i = 0; i <= 3; i++) {      //Shimmy dance
      turnRight();
      delay(100);
      turnLeft();
      delay(100);
    }
    reverse();
    delay(500);
    turnRight();
    delay(1500);
    forward();
  }
}

// Explore mode motion routines
void forward() {
  servoLeft.write(180);
  servoRight.write(0);
}
void forwardSlow() {
  servoLeft.write(100);
  servoRight.write(80);
  turret_enable = false;
}
void reverse() {
  servoLeft.write(0);
  servoRight.write(180);
  turret_enable = false;
}
void turnRight() {
  servoLeft.write(180);
  servoRight.write(180);
  turret_enable = false;
}
void turnLeft() {
  servoLeft.write(0);
  servoRight.write(0);
  turret_enable = false;
}
void stopRobot() {
  servoLeft.write(90);
  servoRight.write(90);
}

// Read Sharp IR (GP2D120) proximity sensor
int irRead() {
  averaging = 0;
  // Get a sampling of 5 readings from sensor
  for (int i=0; i<5; i++) {
    distance = analogRead(A5);
    averaging = averaging + distance;
    delay(50);  //Wait 50 ms between each read
                //According to datasheet time between each read
                // is ~38ms +/- 10ms. Waiting 50 ms assures each
  }             // read is from a different sample
  distance = averaging / 5;      // Average out readings
  if (distance >= NEAR)
    return(1);
  else
    return(0);
}

// Read SRF05 ultrasonic sensor
int sonarRead () {
  averaging = 0;
  for (int i=0; i<5; i++) {           // Read 5 times and average
    pinMode(srfPin, OUTPUT);	      // Set pin to OUTPUT
    digitalWrite(srfPin, LOW);	      // Ensure pin is low
    delayMicroseconds(2);
    digitalWrite(srfPin, HIGH);       // Start ranging 
    delayMicroseconds(10);            //  with 10 microsecond burst
    digitalWrite(srfPin, LOW);        // End ranging
    pinMode(srfPin, INPUT);	      // Set pin to INPUT
    duration = pulseIn(srfPin, HIGH); // Read echo pulse
    distance = duration / 74 / 2;     // Convert to inches
    averaging = averaging + distance;
    delay(1);
  }
  distance = averaging / 5;      // Average out readings
  return(distance);
}

// Turret movement
void turretR() {
  servoTurret.write(20);
  delay(T_DELAY / 2);            // Half delay for 90
}                                //  degreee transit
void turretL() {
  servoTurret.write(160);
  delay(T_DELAY);                // Full delay for 180 
}                                //  degree transit
void turretC() {
  servoTurret.write(90);
  delay(T_DELAY / 2);
}

void showLED() {
  // Show LED if a bumper is hit
  if (pbRight == HIGH || pbLeft == HIGH) {     
    digitalWrite(LED, HIGH);
  } 
  else {
    digitalWrite(LED, LOW); 
  }   
}

//==============================================
// Interrupt handlers
void hitLeft() {
  pbLeft = HIGH;
}
void hitRight() {
  pbRight = HIGH;
}
void turret_scan() {
  mstimer = true;
}  

// Line following methods
//==============================================

// 2-sensor optics
void mode_line_follow() {
  // Read reflective sensors
  irReflectL = analogRead(lineLSense);  
  irReflectR = analogRead(lineRSense);

  if (irReflectL >= thresh && irReflectR >= thresh) {
    line_forward();   // on line
  }
  if (irReflectL >= thresh && irReflectR <= thresh) {
    line_spinLeft();  // veering off right
    delay(4);
  }
  if (irReflectL <= thresh && irReflectR >= thresh) {
    line_spinRight();  // veering off left
    delay(4);
  }
  // If line is lost try to reacquire
  if (irReflectL < thresh && irReflectR < thresh) {
    line_spinRight();
    delay(20);
  }
}

// 5-sensor array (uses Pololu QTR-8RC)
void mode_line_follow_array() {
  unsigned int position = qtrrc.readLine(sensorValues);
    if (position > (2000 - VARIANCE) && position < (2000 + VARIANCE)) {
    line_forward();
  }
  if (position > 2000 + VARIANCE) {
    line_spinLeft();    
    delay(2);
  }
  if (position < 2000 - VARIANCE) {
    line_spinRight();    
    delay(2);
  }
}

// Line following motion routines
void line_forward() {
  servoLeft.write(0);
  servoRight.write(180);
}
void line_spinRight() {
  servoLeft.write(180);
  servoRight.write(180);
}
void line_spinLeft() {
  servoLeft.write(0);
  servoRight.write(0);
}
